% Given: a deterministic automaton A (with all the states final - for the moment)
% Required: a subautomaton A' of  A such that A' is reachable, deadlock-free,
%    all the action of the A are also present in A' and satisfies ID anf FD.
% (some optimisations on the diamonds rules implementation are performed)
% Version 1.2

% choose whether a transition belongs or not to the subautomaton
%% { subaut(Q1,A,Q2) } :- trans(Q1,A,Q2). 
% normal logic programming style 
subaut(Q1,A,Q2) :- trans(Q1,A,Q2), not subaut0(Q1,A,Q2).
subaut0(Q1,A,Q2) :- trans(Q1,A,Q2), not subaut(Q1,A,Q2).


% the states of the initial and sub- automaton
state(Q1) :- trans(Q1,A,Q2).
state(Q2) :- trans(Q1,A,Q2).
% the labels of the initial automaton
action(A) :- trans(Q1,A,Q2).

% [reachability of the subautomaton] 
% >> least fix point procedure!
reach(Q) :- initialstate(Q).
reach(Q2) :- reach(Q1), subaut(Q1,A,Q2), trans(Q1,A,Q2).
:- not reach(Q), subaut(Q,A,Q2), trans(Q,A,Q2).
:- not reach(Q), subaut(Q1,A,Q), trans(Q1,A,Q).

% [each state should be live => deadlock-freedom]
live(Q1) :- subaut(Q1,A,Q2), trans(Q1,A,Q2).
:- not live(Q), reach(Q), state(Q).

% [all action of A are present in A']
subaut_action(A) :- subaut(Q1,A,Q2), trans(Q1,A,Q2).
:- action(A), not subaut_action(A).

% --- diamond rules ID & FD ---
% preliminaries
trans_commute(Q1,A,B) :- indep(A,B), lt(A,B),
    trans(Q1,A,Q2), trans(Q2,B,Q4), trans(Q1,B,Q3), trans(Q3,A,Q4).
subaut_commute(Q1,A,B) :- indep(A,B), lt(A,B),
    subaut(Q1,A,Q2), subaut(Q2,B,Q4), subaut(Q1,B,Q3), subaut(Q3,A,Q4),
    trans(Q1,A,Q2), trans(Q2,B,Q4), trans(Q1,B,Q3), trans(Q3,A,Q4).
% ID
:- indep(A,B), lt(A,B), 
    subaut(Q1,A,Q2), trans(Q1,A,Q2), 
    subaut(Q2,B,Q4), trans(Q2,B,Q4),
    not trans_commute(Q1,A,B).
:- indep(A,B), lt(B,A), 
    subaut(Q1,A,Q2), trans(Q1,A,Q2), 
    subaut(Q2,B,Q4), trans(Q2,B,Q4),
    not trans_commute(Q1,B,A).
:- indep(A,B), lt(A,B),
    subaut(Q1,A,Q2), trans(Q1,A,Q2), 
    subaut(Q2,B,Q4), trans(Q2,B,Q4),
     trans_commute(Q1,A,B), not subaut_commute(Q1,A,B).
:- indep(A,B), lt(B,A), 
    subaut(Q1,A,Q2), trans(Q1,A,Q2), 
    subaut(Q2,B,Q4), trans(Q2,B,Q4),
    trans_commute(Q1,B,A), not subaut_commute(Q1,B,A).
% FD
:- indep(A,B), lt(A,B), 
    subaut(Q1,A,Q2), trans(Q1,A,Q2), 
    subaut(Q1,B,Q3), trans(Q1,B,Q3),
    not trans_commute(Q1,A,B).
:- indep(A,B), lt(A,B), 
    subaut(Q1,A,Q2), trans(Q1,A,Q2), 
    subaut(Q1,B,Q3), trans(Q1,B,Q3),
    trans_commute(Q1,A,B), not subaut_commute(Q1,A,B).

% minimization of the output (optional)
%minimize {subaut(Q1,A,Q2) : trans(Q1,A,Q2)}.
%minimize {reach(Q) : state(Q)}.

